@@ -1,40 +1,18 @@
 # -*- coding: utf-8 -*-
 from django.conf import settings
 from django.contrib.auth.models import User
-from django.core import signals
 from django.test.signals import template_rendered
 from django.core.handlers.wsgi import WSGIHandler
 from django.core.servers.basehttp import AdminMediaHandler
-from django.db import close_connection
-from django.http import HttpResponseServerError
 from django.test import TestCase
 from django.test.client import store_rendered_templates
 from django.utils.functional import curry
-from webtest import TestApp, TestRequest
+from django.utils.importlib import import_module
+from webtest import TestApp
+from webtest.compat import to_string
 
-
-class DjangoWsgiFix(object):
-    """Django closes the database connection after every request;
-    this breaks the use of transactions in your tests. This wraps
-    around Django's WSGI interface and will disable the critical
-    signal handler for every request served.
-
-    Note that we really do need to do this individually a every
-    request, not just once when our WSGI hook is installed, since
-    Django's own test client does the same thing; it would reinstall
-    the signal handler if used in combination with us.
-
-    From django-test-utils.
-    """
-    def __init__(self, app):
-        self.app = app
-
-    def __call__(self, environ, start_response):
-        signals.request_finished.disconnect(close_connection)
-        try:
-            return self.app(environ, start_response)
-        finally:
-            signals.request_finished.connect(close_connection)
+from django_webtest.middleware import DjangoWsgiFix
+from django_webtest.response import DjangoWebtestResponse
 
 
 class DjangoTestApp(TestApp):
@@ -49,12 +27,18 @@ class DjangoTestApp(TestApp):
         if user:
             environ = environ or {}
             if isinstance(user, User):
-                environ['REMOTE_USER'] = str(user.username)
+                environ['WEBTEST_USER'] = to_string(user.username)
             else:
-                environ['REMOTE_USER'] = user
+                environ['WEBTEST_USER'] = to_string(user)
         return environ
 
     def do_request(self, req, status, expect_errors):
+        req.environ.setdefault('REMOTE_ADDR', '127.0.0.1')
+
+        # is this a workaround for https://code.djangoproject.com/ticket/11111 ?
+        req.environ['REMOTE_ADDR'] = to_string(req.environ['REMOTE_ADDR'])
+        req.environ['PATH_INFO'] = to_string(req.environ['PATH_INFO'])
+
         # Curry a data dictionary into an instance of the template renderer
         # callback function.
         data = {}
@@ -83,10 +67,12 @@ class DjangoTestApp(TestApp):
         elif data.get('templates'):
             response.template = flattend('templates')
 
+        response.__class__ = DjangoWebtestResponse
         return response
 
     def get(self, url, params=None, headers=None, extra_environ=None,
-            status=None, expect_errors=False, user=None, auto_follow=False):
+            status=None, expect_errors=False, user=None, auto_follow=False,
+            content_type=None):
         extra_environ = self._update_environ(extra_environ, user)
         response = super(DjangoTestApp, self).get(
                   url, params, headers, extra_environ, status, expect_errors)
@@ -105,6 +91,33 @@ class DjangoTestApp(TestApp):
                    url, params, headers, extra_environ, status,
                    upload_files, expect_errors, content_type)
 
+    def put(self, url, params='', headers=None, extra_environ=None,
+             status=None, upload_files=None, expect_errors=False,
+             content_type=None, user=None):
+        extra_environ = self._update_environ(extra_environ, user)
+        return super(DjangoTestApp, self).put(
+                   url, params, headers, extra_environ, status,
+                   upload_files, expect_errors, content_type)
+
+    def delete(self, url, params='', headers=None, extra_environ=None,
+             status=None, expect_errors=False,
+             content_type=None, user=None):
+        extra_environ = self._update_environ(extra_environ, user)
+        return super(DjangoTestApp, self).delete(
+                   url, params, headers, extra_environ, status,
+                   expect_errors, content_type)
+
+    @property
+    def session(self):
+        """
+        Obtains the current session variables.
+        """
+        if 'django.contrib.sessions' in settings.INSTALLED_APPS:
+            engine = import_module(settings.SESSION_ENGINE)
+            cookie = self.cookies.get(settings.SESSION_COOKIE_NAME, None)
+            if cookie:
+                return engine.SessionStore(cookie)
+        return {}
 
 class WebTest(TestCase):
 
@@ -113,8 +126,9 @@ class WebTest(TestCase):
     setup_auth = True
 
     def _patch_settings(self):
-        ''' Patch settings to add support for REMOTE_USER authorization
-            and (optional) to disable CSRF checks
+        '''
+        Patches settings to add support for django-webtest authorization
+        and (optional) to disable CSRF checks.
         '''
 
         self._DEBUG_PROPAGATE_EXCEPTIONS = settings.DEBUG_PROPAGATE_EXCEPTIONS
@@ -132,38 +146,49 @@ class WebTest(TestCase):
             self._setup_auth()
 
     def _unpatch_settings(self):
-        ''' Restore settings to before-patching state '''
+        ''' Restores settings to before-patching state '''
         settings.MIDDLEWARE_CLASSES = self._MIDDLEWARE_CLASSES
         settings.AUTHENTICATION_BACKENDS = self._AUTHENTICATION_BACKENDS
         settings.DEBUG_PROPAGATE_EXCEPTIONS = self._DEBUG_PROPAGATE_EXCEPTIONS
 
     def _setup_auth(self):
-        ''' Setup REMOTE_USER authorization '''
-        self._setup_remote_user_middleware()
-        self._setup_remote_user_backend()
+        ''' Setups django-webtest authorization '''
+        self._setup_auth_middleware()
+        self._setup_auth_backend()
 
     def _disable_csrf_checks(self):
         disable_csrf_middleware = 'django_webtest.middleware.DisableCSRFCheckMiddleware'
         if not disable_csrf_middleware in settings.MIDDLEWARE_CLASSES:
-            settings.MIDDLEWARE_CLASSES = [disable_csrf_middleware] + settings.MIDDLEWARE_CLASSES
+            settings.MIDDLEWARE_CLASSES.insert(0, disable_csrf_middleware)
 
-    def _setup_remote_user_middleware(self):
-        remote_user_middleware = 'django.contrib.auth.middleware.RemoteUserMiddleware'
-        if not remote_user_middleware in settings.MIDDLEWARE_CLASSES:
-            settings.MIDDLEWARE_CLASSES += [remote_user_middleware]
+    def _setup_auth_middleware(self):
+        webtest_auth_middleware = 'django_webtest.middleware.WebtestUserMiddleware'
+        django_auth_middleware = 'django.contrib.auth.middleware.AuthenticationMiddleware'
 
-    def _setup_remote_user_backend(self):
-        auth_backends = settings.AUTHENTICATION_BACKENDS
-        try:
-            index = auth_backends.index('django.contrib.auth.backends.ModelBackend')
-            auth_backends[index] = 'django.contrib.auth.backends.RemoteUserBackend'
-        except ValueError:
-            auth_backends.append('django.contrib.auth.backends.RemoteUserBackend')
-        settings.AUTHENTICATION_BACKENDS = auth_backends
+        if django_auth_middleware not in settings.MIDDLEWARE_CLASSES:
+            # There can be a custom AuthenticationMiddleware subclass or replacement,
+            # we can't compute its index so just put our auth middleware to the end.
+            # If appending causes problems _setup_auth_middleware method can
+            # be overriden by a subclass.
+            settings.MIDDLEWARE_CLASSES.append(webtest_auth_middleware)
+        else:
+            index = settings.MIDDLEWARE_CLASSES.index(django_auth_middleware)
+            settings.MIDDLEWARE_CLASSES.insert(index+1, webtest_auth_middleware)
+
+    def _setup_auth_backend(self):
+        backend_name = 'django_webtest.backends.WebtestUserBackend'
+        settings.AUTHENTICATION_BACKENDS.insert(0, backend_name)
+
+    def renew_app(self):
+        """
+        Resets self.app (drops the stored state): cookies, etc.
+        Note: this renews only self.app, not the responses fetched by self.app.
+        """
+        self.app = DjangoTestApp(extra_environ=self.extra_environ)
 
     def __call__(self, result=None):
         self._patch_settings()
-        self.app = DjangoTestApp(extra_environ=self.extra_environ)
+        self.renew_app()
         res = super(WebTest, self).__call__(result)
         self._unpatch_settings()
         return res
